(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2020 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`RotatingWaveApproximation`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

Begin["`Private`"];

EffectiveHamiltonian[h_, u_, t_] := ECT[u].h.u - I ECT[u].D[u, t];

DropFastTerms[expr_, freq_, t_] :=
    Module[{freq1, expr1},
        {expr1, freq1} = Expand[
            {expr, EnsureList@freq} //.
                (f : Sin | Cos)[x_ /; ! FreeQ[x, t]] :> TrigToExp@f@ExpandAll@x //.
                    Exp[x_ /; ! FreeQ[x, t]] :> Exp@ExpandAll@x
        ];
        expr1 /. Flatten[{{Exp@ExpandAll[2 I # t + c_.] -> 0, Exp@ExpandAll[-2 I # t + c_.] -> 0} & /@ freq1}]
    ];

Options[RotatingWaveTransformMatrix] = {TimeVariable -> Automatic, ReturnFrequencyReplacements -> False, Method -> Automatic};

OptionProcessor[Method] = EnsureList@#&;

frequencyListPattern[] =
    Except[_List | _Rule] |
        {Except[_Rule], _Rule | {__Rule}} |
        {Except[_Rule], _Rule | {__Rule}, _} |
        {Except[_Rule | _List] ..} |
        {({Except[_Rule]} | {Except[_Rule], _Rule | {__Rule}} | {Except[_Rule], _Rule | {__Rule}, _})..};

standardizeFrequencyList[freqList_] := Switch[freqList,
    Except[_List | _Rule],
    {{freqList}},
    {Except[_Rule | _List] ..},
    List /@ freqList,
    {Except[_Rule], _Rule | {__Rule}} | {Except[_Rule], _Rule | {__Rule}, _},
    {freqList},
    _,
    freqList
];

(RotatingWaveTransformMatrix[sys_?AtomicSystemQ, frequencyList : frequencyListPattern[], OptionsPattern[]] :=
    Module[{timeVariable, method, returnFrequencyReplacements, shiftList, freqReps, freqBackReps, freq,
        freqList = standardizeFrequencyList[frequencyList],
        lowerPos, upperPos, lowerShifts, upperShifts, lowerUnfixed,
        upperUnfixed, Unfixed, transformMatrix, coupling, diffFreq, bareFrequencies},
        freqReps = freqBackReps = {};
        bareFrequencies = Cases[freqList, {f_} -> f];
        If[Length@bareFrequencies > 0,
            (* If frequency is specified with no transition, assume state with zero energy is ground state, shift all others by light frequency. *)
            shiftList = -Chop@Replace[Energy /@ sys /. 0. -> 0, x_ /; x =!= 0 -> First@bareFrequencies, {1}],
            (* Else find shifts to eliminate as many light frequencies as possible in the order of transitions specified. *)
            (* Initialize shiftList *)
            shiftList = ConstantArray[Unfixed, Length[sys]];
            (* Expand multiple couplings in one entry in freqList to multiple entries, and delete third elements *)
            freqList = freqList[[All, ;; 2]] /. {{f_, c : {__Rule}} :> Sequence @@ ({f, #} & /@ c)};
            (* Loop over optical frequencies, defining transform-induced level shifts to eliminate frequencies when possible.
                If not possible, define replacements to write frequencies in terms of frequencies that have been eliminated. *)
            Do[
                freq = freqList[[i, 1]];
                coupling = freqList[[i, 2]];
                (* Find the positions of states specified to be coupled by this frequency *)
                {lowerPos, upperPos} = List @@ (StatePosition[sys, #] & /@ coupling);
                (* Find the previously specified level shifts for these states *)
                {lowerShifts, upperShifts} = Extract[shiftList, #] & /@ {lowerPos, upperPos};
                (* Split out sublists of states that have different previously specified energy shifts*)
                lowerShifts = Transpose[{lowerShifts, lowerPos}];
                upperShifts = Transpose[{upperShifts, upperPos}];
                {lowerShifts, upperShifts} = Split[#, First[#1] === First[#2]&] & /@ {lowerShifts, upperShifts};
                {lowerShifts, lowerPos} = {lowerShifts[[All, All, 1]], lowerShifts[[All, All, 2]]};
                {upperShifts, upperPos} = {upperShifts[[All, All, 1]], upperShifts[[All, All, 2]]};
                (* Find groups of unfixed states. *)
                {lowerUnfixed, upperUnfixed} = Map[AllSameQ[#, Unfixed] &, {lowerShifts, upperShifts}, {2}];
                (* Loop over the sublists of lower and upper states for this frequency, eliminating optical frequencies by shifting states where possible. *)
                Do[
                    Which[
                        (* If neither upper or lower are fixed, and Method is not "ShiftLowerState",
                            fix the lower at its current position and shift the upper down. *)
                        lowerUnfixed[[j]] && upperUnfixed[[k]] && !MemberQ[method, "ShiftLowerState"],
                        shiftList = ReplacePart[shiftList, lowerPos[[j]] -> 0];
                        shiftList = ReplacePart[shiftList, upperPos[[k]] -> -freq],
                        (* If neither are fixed, and Method is "ShiftLowerState", fix the upper at its current position and shift the lower up. *)
                        lowerUnfixed[[j]] && upperUnfixed[[k]] && MemberQ[method, "ShiftLowerState"],
                        shiftList = ReplacePart[shiftList, lowerPos[[j]] -> freq];
                        shiftList = ReplacePart[shiftList, upperPos[[k]] -> 0],
                        (* If the lower is fixed and the upper is unfixed, shift upper to compensate for lower shift, then shift down by optical frequency. *)
                        ! lowerUnfixed[[j]] && upperUnfixed[[k]],
                        shiftList = ReplacePart[shiftList, upperPos[[k]] -> First[lowerShifts[[j]]] - freq],
                        (* If the lower is unfixed and the upper is fixed, shift lower to compensate for upper shift, then shift up by optical frequency. *)
                        lowerUnfixed[[j]] && ! upperUnfixed[[k]],
                        shiftList = ReplacePart[shiftList, lowerPos[[j]] -> First[upperShifts[[k]]] + freq],
                        (* If both are fixed, find replacement rule and back substitution rule
                            for writing the optical frequency in terms of the previously eliminated frequencies and difference frequency. *)
                        ! lowerUnfixed[[j]] && ! upperUnfixed[[k]],
                        diffFreq = Unique[];
                        AppendTo[
                            freqReps,
                            {freq -> -(First@upperShifts[[k]] - First@lowerShifts[[j]]) + diffFreq, {Flatten@lowerPos[[j]], Flatten@upperPos[[k]]}}
                        ];
                        AppendTo[
                            freqBackReps,
                            {diffFreq -> freq + (First@upperShifts[[k]] - First@lowerShifts[[j]]), {Flatten@lowerPos[[j]], Flatten@upperPos[[k]]}}
                        ]
                    ],
                    {j, Length@lowerPos}, {k, Length@upperPos}
                ],
                {i, Length@freqList}
            ]
        ];
        transformMatrix = MatrixExp[I timeVariable DiagonalMatrix[shiftList /. Unfixed -> 0]];
        If[returnFrequencyReplacements, {transformMatrix, freqReps, freqBackReps}, transformMatrix]
    ]) // ProcessOptions;

(RotatingWaveTransformMatrix[sys_?AtomicSystemQ, shiftlist : {__Rule}, OptionsPattern[]] :=
    Module[{reps, shifts, timeVariable},
        reps = (Alternatives @@ SelectStates[sys, #[[1]]] -> #[[2]] &) /@ shiftlist;
        shifts = sys /. reps /. AtomicState[__] -> 0;
        MatrixExp[I timeVariable DiagonalMatrix@shifts]
    ]) // ProcessOptions;

Options[RotatingWaveApproximation] = {TransformMatrix -> Automatic, TimeVariable -> Automatic, Method -> Automatic};

(RotatingWaveApproximation[sys_?AtomicSystemQ, h_?OperatorQ, frequencyList : frequencyListPattern[], opts : OptionsPattern[]] := Module[
    {method, timeVariable, transformMatrix, rotatingHam, freqList = standardizeFrequencyList[frequencyList], freqReps, freqBackReps},
    (* Get transform matrix and replacement rules *)
    transformMatrix = transformMatrix /. Automatic :> RotatingWaveTransformMatrix[sys, freqList, ReturnFrequencyReplacements -> True, FilterRules@opts];
    {transformMatrix, freqReps, freqBackReps} = If[OperatorQ@transformMatrix, {transformMatrix, {}, {}}, transformMatrix];
    (* Apply transform matrix to Hamiltonian *)
    rotatingHam = EffectiveHamiltonian[h, transformMatrix, timeVariable];
    If[!MemberQ[method, "KeepFastTerms"],
        (* Write untransformed optical frequencies in terms of transformed optical frequencies and difference frequencies *)
        rotatingHam = Fold[ReplaceAllAt[#1, #2[[1]], Join[Tuples@#2[[2]], Tuples@Reverse@#2[[2]]]]&, rotatingHam, freqReps];
        (* Drop fast oscillating terms *)
        rotatingHam = DropFastTerms[ExpandAll@rotatingHam, freqList[[All, 1]], timeVariable];
        (* Substitute back for untransformed optical frequencies *)
        rotatingHam = Fold[ReplaceAllAt[#1, #2[[1]], Join[Tuples@#2[[2]], Tuples@Reverse@#2[[2]]]]&, rotatingHam, freqBackReps];
        (* Write all frequencies in terms of detunings, if specified *)
        freqList = freqList /. {x : Except[_List | _Rule], rules : {__Rule}, y___} :> {x, First@rules, y};
        rotatingHam = rotatingHam /. Cases[
            freqList,
            {freq : Except[_List | _Rule], crit1_ -> crit2_, delta : Except[_List | _Rule]} :> (freq -> Energy@SelectState[sys, crit2] - Energy@SelectState[sys, crit1] + delta)
        ]
    ];
    rotatingHam
]) // ProcessOptions;

RotatingWaveApproximation[atomicSystem_?AtomicSystemQ, reg_ExperimentalRegion, h_?OperatorQ, opts : OptionsPattern[]] /; DefinedQ[reg, OpticalParameters] :=
    Module[{opticalParameters = OptionProcessor[OpticalParameters][OpticalParameters@reg], freqList},
            (* Convert optical parameters to a list of {frequency, transition} pairs. *)
            freqList = {standardizeOpticalFieldArguments[Most@First@#][[1, 1]], AllowedCouplings /. Last@# /. AllowedCouplings -> Sequence[]}& /@ opticalParameters;
            RotatingWaveApproximation[atomicSystem, h, freqList, opts]
        ];

RotatingWaveApproximation[as_?AtomicSystemQ, reg_ExperimentalRegion, h_?OperatorQ, OptionsPattern[]] /;
    !DefinedQ[reg, OpticalParameters] := h;

RotatingWaveApproximation[atomicSystem_?AtomicSystemQ, experimentalSystem : {_ExperimentalRegion..}, h : {_?OperatorQ..}, opts : OptionsPattern[]] :=
    MapThread[RotatingWaveApproximation[atomicSystem, #1, #2, opts]&, {experimentalSystem, h}];

End[];

EndPackage[];

